/*
  This file is part of CDO. CDO is a collection of Operators to
  manipulate and analyse Climate model Data.

  Copyright (C) 2003-2020 Uwe Schulzweida, <uwe.schulzweida AT mpimet.mpg.de>
  See COPYING file for copying and redistribution conditions.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; version 2 of the License.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
*/

/*
   This module contains the following operators:

      Intlevel   intlevel        Linear level interpolation
*/

#include <cdi.h>

#include "cdo_options.h"
#include "process_int.h"
#include "cdo_vlist.h"
#include "cdo_zaxis.h"
#include "parse_literals.h"
#include "pmlist.h"
#include "param_conversion.h"


template <typename T>
static double
vert_interp_lev_kernel(float w1, float w2, const T var1L1, const T var1L2, const T missval)
{
  if (DBL_IS_EQUAL(var1L1, missval)) w1 = 0.0f;
  if (DBL_IS_EQUAL(var1L2, missval)) w2 = 0.0f;

  // clang-format off
  if      (IS_EQUAL(w1, 0.0f) && IS_EQUAL(w2, 0.0f)) return missval;
  else if (IS_EQUAL(w1, 0.0f)) return (w2 >= 0.5f) ? var1L2 : missval;
  else if (IS_EQUAL(w2, 0.0f)) return (w1 >= 0.5f) ? var1L1 : missval;
  else                         return var1L1 * (double)w1 + var1L2 * (double)w2;
  // clang-format on
}

//  1D vertical interpolation
template <typename T>
static void
vert_interp_lev(size_t gridsize, T missval, const Varray<T> &vardata1, Varray<T> &vardata2, int nlev2,
                const Varray<int> &lev_idx1, const Varray<int> &lev_idx2, const Varray<float> &lev_wgt1, const Varray<float> &lev_wgt2)
{
  for (int ilev = 0; ilev < nlev2; ++ilev)
    {
      const auto idx1 = lev_idx1[ilev];
      const auto idx2 = lev_idx2[ilev];
      auto wgt1 = lev_wgt1[ilev];
      auto wgt2 = lev_wgt2[ilev];
      auto var2 = &vardata2[gridsize * ilev];

      auto var1L1 = &vardata1[gridsize * idx1];
      auto var1L2 = &vardata1[gridsize * idx2];

#ifdef _OPENMP
#pragma omp parallel for default(none) shared(gridsize, var2, var1L1, var1L2, missval, wgt1, wgt2)
#endif
      for (size_t i = 0; i < gridsize; ++i)
        {
          var2[i] = vert_interp_lev_kernel(wgt1, wgt2, var1L1[i], var1L2[i], missval);
        }
    }
}

static void
vert_interp_lev(size_t gridsize, double missval, const Field3D &field1, Field3D &field2, int nlev2,
                const Varray<int> &lev_idx1, const Varray<int> &lev_idx2, const Varray<float> &lev_wgt1, const Varray<float> &lev_wgt2)
{
  if (field1.memType == MemType::Float)
    {
      vert_interp_lev(gridsize, (float)missval, field1.vec_f, field2.vec_f, nlev2, lev_idx1, lev_idx2, lev_wgt1, lev_wgt2);
    }
  else
    {
      vert_interp_lev(gridsize, missval, field1.vec_d, field2.vec_d, nlev2, lev_idx1, lev_idx2, lev_wgt1, lev_wgt2);
    }
}

//  3D vertical interpolation
template <typename T>
void
vert_interp_lev3d(size_t gridsize, T missval, const Varray<T> &vardata1, Varray<T> &vardata2, int nlev2,
                  const Varray<int> &lev_idx1, const Varray<int> &lev_idx2, const Varray<float> &lev_wgt1, const Varray<float> &lev_wgt2)
{
  for (int ilev = 0; ilev < nlev2; ilev++)
    {
      auto offset = ilev * gridsize;
      auto var2 = &vardata2[offset];

#ifdef _OPENMP
#pragma omp parallel for default(none) shared(gridsize, offset, vardata1, var2, lev_idx1, lev_idx2, lev_wgt1, lev_wgt2, missval)
#endif
      for (size_t i = 0; i < gridsize; i++)
        {
          const auto idx1 = lev_idx1[offset + i];
          const auto idx2 = lev_idx2[offset + i];
          const auto wgt1 = lev_wgt1[offset + i];
          const auto wgt2 = lev_wgt2[offset + i];

          // upper/lower values from input field
          const auto var1L1 = vardata1[idx1];
          const auto var1L2 = vardata1[idx2];

          var2[i] = vert_interp_lev_kernel(wgt1, wgt2, var1L1, var1L2, missval);
        }
    }
}

// Explicit instantiation
template void
vert_interp_lev3d(size_t gridsize, float missval, const Varray<float> &vardata1, Varray<float> &vardata2, int nlev2,
                  const Varray<int> &lev_idx1, const Varray<int> &lev_idx2, const Varray<float> &lev_wgt1, const Varray<float> &lev_wgt2);
template void
vert_interp_lev3d(size_t gridsize, double missval, const Varray<double> &vardata1, Varray<double> &vardata2, int nlev2,
                  const Varray<int> &lev_idx1, const Varray<int> &lev_idx2, const Varray<float> &lev_wgt1, const Varray<float> &lev_wgt2);

void
vert_interp_lev3d(size_t gridsize, double missval, const Field3D &field1, Field3D &field2, int nlev2,
                  const Varray<int> &lev_idx1, const Varray<int> &lev_idx2, const Varray<float> &lev_wgt1, const Varray<float> &lev_wgt2)
{
  if (field1.memType == MemType::Float)
    {
      vert_interp_lev3d(gridsize, (float)missval, field1.vec_f, field2.vec_f, nlev2, lev_idx1, lev_idx2, lev_wgt1, lev_wgt2);
    }
  else
    {
      vert_interp_lev3d(gridsize, missval, field1.vec_d, field2.vec_d, nlev2, lev_idx1, lev_idx2, lev_wgt1, lev_wgt2);
    }
}

void
vert_gen_weights(int expol, int nlev1, const Varray<double> &lev1, int nlev2, const Varray<double> &lev2, Varray<int> &lev_idx1, Varray<int> &lev_idx2,
                 Varray<float> &lev_wgt1, Varray<float> &lev_wgt2)
{
  for (int i2 = 0; i2 < nlev2; ++i2)
    {
      int idx1 = 0, idx2 = 0;
      double val1, val2 = 0.0;

      // Because 2 levels were added to the source vertical coordinate (one on top, one at the bottom), its loop starts at 1
      int i1;
      for (i1 = 1; i1 < nlev1; ++i1)
        {
          const auto lev1_isUp = (lev1[i1 - 1] < lev1[i1]);
          idx1 = lev1_isUp ? i1 - 1 : i1;
          idx2 = lev1_isUp ? i1 : i1 - 1;
          val1 = lev1[idx1];
          val2 = lev1[idx2];
          if (lev2[i2] > val1 && lev2[i2] <= val2) break;
        }

      if (i1 == nlev1) cdoAbort("Level %g not found!", lev2[i2]);

      if (i1 - 1 == 0)  // destination levels is not covert by the first two input z levels
        {
          lev_idx1[i2] = 1;
          lev_idx2[i2] = 1;
          lev_wgt1[i2] = 0.0f;
          lev_wgt2[i2] = (expol || IS_EQUAL(lev2[i2], val2));
        }
      else if (i1 == nlev1 - 1)  // destination level is beyond the last value of the input z field
        {
          lev_idx1[i2] = nlev1 - 2;
          lev_idx2[i2] = nlev1 - 2;
          lev_wgt1[i2] = (expol || IS_EQUAL(lev2[i2], val2));
          lev_wgt2[i2] = 0.0f;
        }
      else  // target z values has two bounday values in input z field
        {
          lev_idx1[i2] = idx1;
          lev_idx2[i2] = idx2;
          lev_wgt1[i2] = (lev1[idx2] - lev2[i2]) / (lev1[idx2] - lev1[idx1]);
          lev_wgt2[i2] = (lev2[i2] - lev1[idx1]) / (lev1[idx2] - lev1[idx1]);
        }
      // backshift of the indices because of the two additional levels in input vertical coordinate
      lev_idx1[i2]--;
      lev_idx2[i2]--;
      /*
        printf("%d %g %d %d %g %g %d %d %g %g\n", i2, lev2[i2], idx1, idx2, lev1[idx1], lev1[idx2],
        lev_idx1[i2], lev_idx2[i2], lev_wgt1[i2], lev_wgt2[i2]);
      */
    }
}

bool
levelDirUp(const int nlev, const double *const lev)
{
  const bool lup = (nlev > 1 && lev[1] > lev[0]);
  for (int k = 1; k < nlev - 1; ++k)
    if (lup && lev[k + 1] <= lev[k]) return false;

  return lup;
}

bool
levelDirDown(const int nlev, const double *const lev)
{
  const bool ldown = (nlev > 1 && lev[1] < lev[0]);
  for (int k = 1; k < nlev - 1; ++k)
    if (ldown && lev[k + 1] >= lev[k]) return false;

  return ldown;
}

template <typename T>
static void
vert_gen_weights3d1d(bool expol, size_t gridsize, int nlev1, const Varray<T> &xlev1, int nlev2, const Varray<double> &lev2,
                     Varray<int> &xlev_idx1, Varray<int> &xlev_idx2, Varray<float> &xlev_wgt1, Varray<float> &xlev_wgt2)
{
  Varray<double> lev1p2(nlev1 + 2);
  Varray<float> lev_wgt1(nlev2), lev_wgt2(nlev2);
  Varray<int> lev_idx1(nlev2), lev_idx2(nlev2);

  // Check monotony of vertical levels
  for (int k = 0; k < nlev1; ++k) lev1p2[k] = xlev1[k * gridsize];
  const auto lup = levelDirUp(nlev1, lev1p2.data());
  const auto ldown = levelDirDown(nlev1, lev1p2.data());
  if (!lup && !ldown) cdoAbort("Non monotonic zaxis!");
  const double level_0 = lup ? -1.e33 : 1.e33;
  const double level_N = lup ? 1.e33 : -1.e33;

  for (size_t i = 0; i < gridsize; i++)
    {
      lev1p2[0] = level_0;
      lev1p2[nlev1 + 1] = level_N;
      for (int k = 0; k < nlev1; ++k) lev1p2[k + 1] = xlev1[k * gridsize + i];

      vert_gen_weights(expol, nlev1 + 2, lev1p2, nlev2, lev2, lev_idx1, lev_idx2, lev_wgt1, lev_wgt2);

      for (int k = 0; k < nlev2; ++k) xlev_idx1[k * gridsize + i] = lev_idx1[k] * gridsize + i;
      for (int k = 0; k < nlev2; ++k) xlev_idx2[k * gridsize + i] = lev_idx2[k] * gridsize + i;
      for (int k = 0; k < nlev2; ++k) xlev_wgt1[k * gridsize + i] = lev_wgt1[k];
      for (int k = 0; k < nlev2; ++k) xlev_wgt2[k * gridsize + i] = lev_wgt2[k];
    }
}

static void
vert_gen_weights3d1d(bool expol, size_t gridsize, int nlev1, Field3D &field1, int nlev2, const Varray<double> &lev2,
                     Varray<int> &lev_idx1, Varray<int> &lev_idx2, Varray<float> &lev_wgt1, Varray<float> &lev_wgt2)
{
  if (field1.memType == MemType::Float)
    {
      vert_gen_weights3d1d(expol, gridsize, nlev1, field1.vec_f, nlev2, lev2, lev_idx1, lev_idx2, lev_wgt1, lev_wgt2);
    }
  else
    {
      vert_gen_weights3d1d(expol, gridsize, nlev1, field1.vec_d, nlev2, lev2, lev_idx1, lev_idx2, lev_wgt1, lev_wgt2);
    }
}


static void
intlevelGetParameter(Varray<double> &lev2, std::string &zdescription, std::string &zvarname)
{
  const auto pargc = operatorArgc();
  if (pargc)
    {
      const auto pargv = cdoGetOperArgv();

      KVList kvlist;
      kvlist.name = "INTLEVEL";
      if (kvlist.parseArguments(pargc, pargv) != 0) cdoAbort("Parse error!");
      if (Options::cdoVerbose) kvlist.print();

      bool def_level = false;
      bool def_zdescription = false;
      for (const auto &kv : kvlist)
        {
          const auto &key = kv.key;
          //if (kv.nvalues > 1) cdoAbort("Too many values for parameter key >%s<!", key.c_str());
          if (kv.nvalues < 1) cdoAbort("Missing value for parameter key >%s<!", key.c_str());
          const auto &values = kv.values;
          const auto &value = kv.values[0];
          int nvalues = kv.nvalues;
          if (nvalues == 1 && value.empty()) nvalues = 0;

          // clang-format off
         if (key == "level")
            {
              if (def_zdescription) cdoAbort("Parameter level and zdescription can't be mixed!");
              lev2.resize(nvalues);
              for (int i = 0; i < nvalues; ++i) lev2[i] = literal2double(values[i]);
              def_level = true;
            }
          else if (key == "zdescription")
            {
              if (def_level) cdoAbort("Parameter zdescription and level can't be mixed!");
              zdescription = value;
              def_zdescription = true;
            }
          else if (key == "zvarname")
            {
              zvarname = value;
            }
          else cdoAbort("Invalid parameter key >%s<!", key.c_str());
          // clang-format on
        }
    }
}

void *
Intlevel(void *process)
{
  int nrecs;
  int varID, levelID;
  int zaxisID1 = -1;
  int nlevel = 0;

  cdoInitialize(process);

  // clang-format off
  const auto INTLEVEL  = cdoOperatorAdd("intlevel",  0, 0, nullptr);
  const auto INTLEVELX = cdoOperatorAdd("intlevelx", 0, 0, nullptr);
  // clang-format on

  (void) (INTLEVEL);  // CDO_UNUSED

  const auto operatorID = cdoOperatorID();

  const bool expol = (operatorID == INTLEVELX);

  operatorInputArg("level|zdescription, [zvarname]");

  Varray<double> lev2;
  int zaxisID2 = CDI_UNDEFID;
  std::string zvarname;
  auto argv = cdoGetOperArgv();

  if (isdigit((int) argv[0][0]))
    {
      lev2 = cdoArgvToFlt(argv);
    }
  else
    {
      std::string zdescription;
      intlevelGetParameter(lev2, zdescription, zvarname);
      if (!zdescription.empty())
        {
          auto zfilename = zdescription.c_str();
          auto zfp = fopen(zfilename, "r");
          if (!zfp) cdoAbort("Open failed on %s", zfilename);
          zaxisID2 = zaxisFromFile(zfp, zfilename);
          fclose(zfp);
          if (zaxisID2 == CDI_UNDEFID) cdoAbort("Invalid zaxis description file %s!", zfilename);
          const auto nlevels = zaxisInqSize(zaxisID2);
          lev2.resize(nlevels);
          zaxisInqLevels(zaxisID2, lev2.data());            
        }
    }

  const int nlev2 = lev2.size();

  if (Options::cdoVerbose)
    for (int i = 0; i < nlev2; ++i) cdoPrint("lev2 %d: %g", i, lev2[i]);

  const auto streamID1 = cdoOpenRead(0);

  const auto vlistID1 = cdoStreamInqVlist(streamID1);
  const auto vlistID2 = vlistDuplicate(vlistID1);

  const auto taxisID1 = vlistInqTaxis(vlistID1);
  const auto taxisID2 = taxisDuplicate(taxisID1);
  vlistDefTaxis(vlistID2, taxisID2);

  VarList varList1;
  varListInit(varList1, vlistID1);
  varListSetUniqueMemtype(varList1);
  const auto memtype = varList1[0].memType;

  const auto nvars = vlistNvars(vlistID1);

  // Find z-variable
  int nlev1 = 0;
  Varray<double> lev1;
  int zvarID = CDI_UNDEFID;
  bool zvarIsVarying = false;
  size_t zvarGridsize = 0;
  size_t wisize = 0;
  if (!zvarname.empty())
    {
      for (varID = 0; varID < nvars; varID++)
        {
          if (cstrIsEqual(zvarname.c_str(), varList1[varID].name))
            {
              zvarID = varID;
              break;
            }
        }

      if (zvarID == CDI_UNDEFID) cdoAbort("Variable %s not found!", zvarname.c_str());
      zvarIsVarying = varList1[zvarID].timetype == TIME_VARYING;
      zvarGridsize = varList1[zvarID].gridsize;
      nlev1 = varList1[zvarID].nlevels;

      if (zaxisID2 == CDI_UNDEFID)
        {
          zaxisID2 = zaxisCreate(ZAXIS_GENERIC, nlev2);

          char str[CDI_MAX_NAME];
          strcpy(str, "zlev");
          cdiDefKeyString(zaxisID2, CDI_GLOBAL, CDI_KEY_NAME, str);
          str[0] = 0;
          vlistInqVarLongname(vlistID1, zvarID, str);
          if (str[0]) cdiDefKeyString(zaxisID2, CDI_GLOBAL, CDI_KEY_LONGNAME, str);
          str[0] = 0;
          vlistInqVarUnits(vlistID1, zvarID, str);
          if (str[0]) cdiDefKeyString(zaxisID2, CDI_GLOBAL, CDI_KEY_UNITS, str);

          zaxisDefLevels(zaxisID2, lev2.data());
        }

      wisize = zvarGridsize * nlev2;

      int i;
      const auto nzaxis = vlistNzaxis(vlistID1);
      for (i = 0; i < nzaxis; i++)
        {
          const auto zaxisID = vlistZaxis(vlistID1, i);
          nlevel = zaxisInqSize(zaxisID);
          if (nlevel == nlev1)
            {
              zaxisID1 = zaxisID;
              break;
            }
        }
      if (i == nzaxis) cdoAbort("No processable variable found!");
    }
  else
    {
      int i;
      const auto nzaxis = vlistNzaxis(vlistID1);
      for (i = 0; i < nzaxis; i++)
        {
          const auto zaxisID = vlistZaxis(vlistID1, i);
          nlevel = zaxisInqSize(zaxisID);
          //if (zaxisInqType(zaxisID) != ZAXIS_HYBRID && zaxisInqType(zaxisID) != ZAXIS_HYBRID_HALF)
            if (nlevel > 1)
              {
                zaxisID1 = zaxisID;
                break;
              }
        }
      if (i == nzaxis) cdoAbort("No processable variable found!");

      if (zaxisID2 == CDI_UNDEFID)
        {
          zaxisID2 = zaxisCreate(zaxisInqType(zaxisID1), nlev2);

          char str[CDI_MAX_NAME];
          int length = CDI_MAX_NAME;
          cdiInqKeyString(zaxisID1, CDI_GLOBAL, CDI_KEY_NAME, str, &length);
          cdiDefKeyString(zaxisID2, CDI_GLOBAL, CDI_KEY_NAME, str);
          str[0] = 0;
          length = CDI_MAX_NAME;
          cdiInqKeyString(zaxisID1, CDI_GLOBAL, CDI_KEY_LONGNAME, str, &length);
          if (str[0]) cdiDefKeyString(zaxisID2, CDI_GLOBAL, CDI_KEY_LONGNAME, str);
          str[0] = 0;
          length = CDI_MAX_NAME;
          cdiInqKeyString(zaxisID1, CDI_GLOBAL, CDI_KEY_UNITS, str, &length);
          if (str[0]) cdiDefKeyString(zaxisID2, CDI_GLOBAL, CDI_KEY_UNITS, str);

          zaxisDefDatatype(zaxisID2, zaxisInqDatatype(zaxisID1));

          zaxisDefLevels(zaxisID2, lev2.data());
        }

      nlev1 = nlevel;
      lev1.resize(nlev1 + 2);
      cdoZaxisInqLevels(zaxisID1, &lev1[1]);

      const auto lup = levelDirUp(nlev1, &lev1[1]);
      const auto ldown = levelDirDown(nlev1, &lev1[1]);
      if (!lup && !ldown) cdoAbort("Non monotonic zaxis!");
      lev1[0] = lup ? -1.e33 : 1.e33;
      lev1[nlev1 + 1] = lup ? 1.e33 : -1.e33;

      if (Options::cdoVerbose)
        for (i = 0; i < nlev1 + 2; ++i) cdoPrint("lev1 %d: %g", i, lev1[i]);

      wisize = nlev2;
    }

  const auto nzaxis = vlistNzaxis(vlistID1);
  for (int index = 0; index < nzaxis; index++)
    {
      const auto zaxisID = vlistZaxis(vlistID1, index);
      const auto nlevels = zaxisInqSize(zaxisID);
      if (zaxisID == zaxisID1 || nlevels == nlev1)
        vlistChangeZaxisIndex(vlistID2, index, zaxisID2);
    }

  VarList varList2;
  varListInit(varList2, vlistID2);
  varListSetMemtype(varList2, memtype);

  Varray<int> lev_idx1(wisize), lev_idx2(wisize);
  Varray<float> lev_wgt1(wisize), lev_wgt2(wisize);

  const auto streamID2 = cdoOpenWrite(1);

  cdoDefVlist(streamID2, vlistID2);

  std::vector<bool> vars(nvars);
  std::vector<bool> varinterp(nvars);
  std::vector<std::vector<size_t>> varnmiss(nvars);
  Field3DVector vardata1(nvars), vardata2(nvars);

  const int maxlev = (nlev1 > nlev2) ? nlev1 : nlev2;

  for (varID = 0; varID < nvars; varID++)
    {
      const auto nlevels = varList1[varID].nlevels;

      vardata1[varID].init(varList1[varID]);

      varinterp[varID] = (varList1[varID].zaxisID == zaxisID1 || nlevels == nlev1);

      if (varinterp[varID])
        {
          varnmiss[varID].resize(maxlev, 0);
          vardata2[varID].init(varList2[varID]);
        }
      else
        {
          varnmiss[varID].resize(nlevels);
        }
    }

  int tsID = 0;
  while ((nrecs = cdoStreamInqTimestep(streamID1, tsID)))
    {
      for (varID = 0; varID < nvars; ++varID) vars[varID] = false;

      taxisCopyTimestep(taxisID2, taxisID1);
      cdoDefTimestep(streamID2, tsID);

      for (int recID = 0; recID < nrecs; recID++)
        {
          cdoInqRecord(streamID1, &varID, &levelID);
          cdoReadRecord(streamID1, vardata1[varID], levelID, &varnmiss[varID][levelID]);
          vars[varID] = true;
        }

      if (tsID == 0 || zvarIsVarying)
        {
          if (!zvarname.empty())
            vert_gen_weights3d1d(expol, zvarGridsize, nlev1, vardata1[zvarID], nlev2, lev2, lev_idx1, lev_idx2, lev_wgt1, lev_wgt2);
          else
            vert_gen_weights(expol, nlev1 + 2, lev1, nlev2, lev2, lev_idx1, lev_idx2, lev_wgt1, lev_wgt2);
        }

      for (varID = 0; varID < nvars; varID++)
        {
          if (vars[varID] && varinterp[varID])
            {
              const auto missval = varList1[varID].missval;
              const auto gridsize = varList1[varID].gridsize;

              if (!zvarname.empty())
                vert_interp_lev3d(gridsize, missval, vardata1[varID], vardata2[varID], nlev2, lev_idx1, lev_idx2, lev_wgt1, lev_wgt2);
              else
                vert_interp_lev(gridsize, missval, vardata1[varID], vardata2[varID], nlev2, lev_idx1, lev_idx2, lev_wgt1, lev_wgt2);

              for (levelID = 0; levelID < nlev2; levelID++)
                {
                  const auto offset = gridsize * levelID;
                  if (memtype == MemType::Float)
                    varnmiss[varID][levelID] = arrayNumMV(gridsize, &vardata2[varID].vec_f[offset], (float)missval);
                  else
                    varnmiss[varID][levelID] = arrayNumMV(gridsize, &vardata2[varID].vec_d[offset], missval);
                }
            }
        }

      for (varID = 0; varID < nvars; varID++)
        {
          if (vars[varID])
            {
              for (levelID = 0; levelID < varList2[varID].nlevels; levelID++)
                {
                  cdoDefRecord(streamID2, varID, levelID);
                  cdoWriteRecord(streamID2, varinterp[varID] ? vardata2[varID] : vardata1[varID], levelID, varnmiss[varID][levelID]);
                }
            }
        }

      tsID++;
    }

  cdoStreamClose(streamID2);
  cdoStreamClose(streamID1);

  cdoFinish();

  return nullptr;
}
